* SBasic-Funktionen QSEND, QKEYQ, KEYQ fr QL-Today von .hpr'96
* Copyright (C) 1996 H.-Peter Recktenwald, Berlin

doctext MACRO ; ende des textteils bei ENDM

* i/o-exchange between SBasic programs *

- "SBasic" for the QL's Super-BASIC;
- "PIF" for Pointer InterFace (as from "ptr_gen").


"Signals" to another program?
UNIX, or LINUX, the ultimate system? Maybe, who cares! Why not
using some of its features for the QL (which anyway is the case,
and also the opposite way).

There is litte use in mass production of monstrous "ports" of
other programs from a strange world. Those didn't solve many
tasks sufficiently, or, if so, at the price of such huge memory
requirements, that their use would become quite expensive, and
not adequate to the true capabilities of the QLs operating system.

There are many ways to go, and one of the most easiest and even
very safe ones is accessing another jobs i/o-queue by simulating
the keybord input to that one from any other job. Not really
those "signals" but, of much more versatility and practical use.

* ++

len = QSEND([#]channel,Text$)

Sends some text (any code in the range of 0 to 255 may be used)
to the input queue of a channel. If that channel is the i/o-channel
of another job the text can be used to control every feature of it.

QSEND by default has no timeout, it will send whatever the receiver
accepts at the first attempt, then return. This can be modified with
the delimiter being "!", to an endless timeout, returning only after
all charcters have been sent.
Alternatively the "TO" separator can be used to additionally pick
the job just before transferring the string data. - Which will
only work if the calling jobs channel #0 is an accessible window.

The proceedings can always be interrupted with the <break> keys.

A PIPE represented by [#]channel additionally will be modified to
a bi-directional device, that can be PRINTed to and can send its
contents to INPUT or INKEY$ commands.

        len = QSEND(#pipechan,'')
will just do the modification, leaving th PIPE to further use.
        len = QSEND(channel TO '')
will just pick (activate) the owner-job of channel (needs the PIF).

        Parameters:

#channel        the SBasic channel number;
channel         the QDOS channel-ID or,
                just the QDOS channel number

text$           any string of any (reasonable) length.

        Function return:

len > LEN(Text$) if all text was sent;
len > 0         the position of where to continue with
                the next call to QSEND to send the
                remaining characters;
 0              if nothing was sent;
negative        if a QDOS error occurred.


channel = KEYQ(jobnumber)

Finds the system number (the "QDOS-channel") of a jobs 1st CON-channel,
which usually is be keyboard i/o-channel.

jobnumber       can be passed as the full job-ID or the QDOS-number of
                the job concerned, as displayed by QJ (QLib-extensions)
                or by JOBS (TK2).


channel = QKEYQ

As KEYQ, with no jobnumber, referring to the currently active job.
This is (mostly) more reliable as KEYQ, it works even if the PIF
has taken over the i/o-queue (after the pointer sprite was visible).


ernum = PCONNECT([#]sending_channel,[#]receiving_channel)

Connecting two PIPE channels together or, a PIPE sender with a CON
receiver. The call works with endless timeout, it can only be
interrupted with the <break> keys, when called from an interpreting
SBasic job.
PCONNECT is sometimes the more adequate solution, and often faster
than QSEND.

* ++

An example, controlling a QUILL job. It not only enables complete
outer control of the job, entering texts or commands, while leving
the standard keyboard input intact as it is. It also works as an
input filter to safely enter corrupted and otherwise lost text or
doc-files, which could lock the QUILL-job due to false characters
when trying to fetch them regularly or by an "import" command.

JOBS or QJ will display the required job number.
        q = KEYQ(jobnumber)

Alternatively, QKEYQ can be easier to use, and more reliable:
        PAUSE 200 : q = QKEYQ

Entering the above line leaves a little less than four seconds
to activating the job concerned, by CTR/C or else, and after
the PAUSE elapsed the variable will contain the required channel
number.

Some QUILL commands are more handy to send if defined as
special strings:
        F3$ = CHR$(240)
        F2$ = CHR$(236)
        Nl$ = CHR$(10)

A trial text:
        text$ = F2$&F3$&'lflp2_test'&Nl$

Fetching the jobs i/o-channel
        q = KEYQ ( jobnumber )
or enter
        PAUSE 200:q=QKEYQ:BEEP 999,99
then activate the QUILL job, and after the signal
go back to SBasic and proceed.

Passing the text:
        s=1 : t=LEN( Text$ )
        REPeat r
          l = QSEND( q , Text$(s TO) )
          s=s+l : IF l < 0 OR s > t : EXIT r
        END REPeat r
        IF l < 0 : PRINT"Error:"!!:REPORT#1,l

Which, if succeding, will adjust the QUILL display to full screen
and load the document "flp2_test_doc".
If QSEND was called with the "!" delimiter that one line would
suffice instead of the REPeat loop:
          l = QSEND( q ! Text$ )

The PCONNECT alternative (often much faster operating):
        p = FOPEN( "PIPE_200" )
        q = KEYQ ( jobnumber )
        r = PCONNECT( #p , q )
        IF r > 0 : PRINT#p,Text$


That is it, for now, much success...

* ++

The extension code:

The extension can be loaded and activated any time by RESPR(862),
LBYTES and CALL or, by LRESPR filename.

It was tested with a GoldCard extended QL, JM rsp. Minerva(1.93) roms
fitted, a 512K internal memory extended standard QL with Minerva,
and in the QXL system with SMSQ/E 2.77. It does NOT depend on the PIF.

Some (elder?) Atari E-software, or other, depending on the PIF
version actually in use, might need the cell at address 4 of the
code (a 16-bit value) being patched to zero. This cannot be determined
from within the code without involving quite lengthy extra routines
that were left off to keeping the code at a (fairly) handy size.
If, for example, KEYQ would return only meaningless values or just
error numbers this modification might cure the problem.

* ++

The assembly was made to suit the QUANTA-GST-Macro-Assembler.

Alternatively the program code can be produced with the following
procedure and the DATA statements listed below.
Enter the procedure and the DATA lines, and call DAFI, with no
parameters. This will produce the code ready to LRESPR (etc.)
as "ram1_qsend_byt", the name and destintion device of which
may be changed to suit by editting the file name at line 10010.


1680 :
1690 DEFine PROCedure DAFI
1700 h$='123456789abcdef'
1710 RESTORE 1710:READ f$,e$,f
1720 DELETE f$:OPEN_NEW#4,f$
1730 c=0:n=0:m=0:l%=0
1740 REPeat l
1750  c%=0:l%=l%+1:READ e$:IF e$='#':EXIT l
1760  FOR g=1TO LEN(e$)STEP 2
1790   e%=16*(e$(g)INSTR h$)+e$(g+1)INSTR h$
1800   c%=c%+e%:n=n+1:IF e%:m=n
1810   PRINT#4,CHR$(e%);
1880  END FOR g
1890  c=c+c%:READ r
1900  IF r<>c%:PRINT'Error in line'!l%!':'!c%!'=>'!r
1910 END REPeat l
1920 READ r
1930 CLOSE#4
1940 PRINT'Checksum'!c,l%!'lines'\f$,n!'bytes.'
1960 IF f<m OR f>n:PRINT'Length error, should be'!f!'bytes.'
1970 IF c<>r:PRINT'Global checksum error, should be'!r;'.'
1980 END DEFine
1990 :
10000 :
10010 DATA 'ram1_qsend_byt','hex',862
10020 DATA '60048000003043FA000C347801104E92',1018
10030 DATA '70004E7500000000000402B4044B4559', 730
10040 DATA '510002100850434F4E4E454354000018', 733
10050 DATA '05514B45595100A2055153454E440000', 946
10060 DATA '000450495045617E246C004C74FF206C',1260
10070 DATA '00782A502A6D0004382C00723004E548', 964
10080 DATA '41F0000420206B142240E2186B0EB489',1286
10090 DATA '6714B5C96B062649D7D3B7CA5ACCFFE6',2319
10100 DATA '6B2ABBE90004670000C478FE4A826E1C',1588
10110 DATA '76504403B6916214D2FC00184A99660C',1541
10120 DATA '4A996608343AFF5ED2C24A916600009E',1679
10130 DATA '246D00904A2A00116B06246AFFF8D4C2',1586
10140 DATA '94C2240A608872023278011A4E917000',1268
10150 DATA '4E4128482A014E7561EC61000180666C',1262
10160 DATA 'D402EA0A590257C7530257C2D4028E02',1559
10170 DATA '616E66582C2900086100010C664670F4',1128
10180 DATA '3831E8006B3C52076A1622062A08206E', 953
10190 DATA '00302070E80074027600707C4E432045',1142
10200 DATA '387800E03404600852891231E8014E94',1305
10210 DATA '5BCAFFF6670C4A076E0853894AEE008F',1783
10220 DATA '66E69842D1EE005C2D480058226E0058',1526
10230 DATA '55892D4900583384E800780370004E75',1273
10240 DATA '2041224A700945FAFFF476004E432449',1516
10250 DATA '46806E5470704E43263AFE8822084680',1487
10260 DATA '6F027600206C0078225820690004B1EA',1165
10270 DATA '0004224A66324A83671AD4C3D5E8006C',1558
10280 DATA '240A6F20E21A6B1C24126F18588A598A',1218
10290 DATA '78004E752052D1CAD4FC006CB1D267EE',2140
10300 DATA 'B1E267EA78F94E75260141FAFE7472FF',2397
10310 DATA '70014E422203280066EC70024E4241EA',1229
10320 DATA '001C201059886E04201867D84A906E02',1120
10330 DATA '208076FF224A244078004E7532780118',1251
10340 DATA '1233E80160043278011670F1206E0058',1178
10350 DATA '91EE005C7608D68BBBC3650E48E742A4',1984
10360 DATA '2A434E91264D4CDF2542280066025343',1143
10370 DATA '4E756100FE926124661E6100FF246618',1471
10380 DATA '78F44A436A127C1CDC896110660A6100',1460
10390 DATA 'FF1066042246228A6000FEF2619E6B54',1691
10400 DATA '2E3698006B4ED1EE005C2D48005878F9',1550
10410 DATA '14016A14CFFC00286B3ADEAE0030BEAE',1619
10420 DATA '003464302E36780022076B28BE6C0072',1020
10430 DATA '622278FA246C0078E549223210006B14',1295
10440 DATA '2441222A00103207B287670642474A87',1018
10450 DATA '660278004E7578FEBE6C006262F67200',1647
10460 DATA '32076A0422052E05246C006860CA6100', 900
10470 DATA 'FDF66100FF286600FE6C2E31E8006A02',1790
10480 DATA '2E05D1EE005C2D48005861CA6600FE5E',1544
10490 DATA '206C0078225026290004382C00723004', 723
10500 DATA 'E54841F0000420206B162240E2986B10',1402
10510 DATA 'B2A90008660AB6A90004660424493404',1093
10520 DATA '51CCFFE43802222A0010320220417070',1291
10530 DATA '76004E4346806E22D4FAFCCA4A2A0011',1654
10540 DATA '6B18202AFFF82440E280630E6B0C282A',1476
10550 DATA '0010380A986C007AE44C6000FDF0',1357
10560 DATA '#',74236
20000 :

The Assembly:

'cdt'   channel definition table
'jdt'   job definition table, job number as index to that table
'q.chan' system number of a channel, index to table of cdt's
'PIF'   pointer interface

 ENDM

        pagelen 20000
        data 0
        section qs

* link SB-extensions to the system
start   bra.s   go
piflag  dc.w    $8000           -ve for flag "PIF present" (dont't change!)
pifdsp  dc.w    48              probably must be zero with Atari E-software
go      lea     bastab(pc),a1
        move.w  $110,a2         bp.init
        jsr     (a2)            > link extension(s) to basic(-job)
        moveq   #0,d0           no error (might help with certain QL-roms)
        rts

bastab  dc.w 0,0,4              no procedures, four functions
        dc.w keyq-*
        dc.b 4,'KEYQ'           q.channel of a jobs i/o-window
        ds.w 0                  adjust to even address
        dc.w pconnect-*
        dc.b 8,'PCONNECT'       connect pipe to pipe or window
        ds.w 0
        dc.w qkeyq-*
        dc.b 5,'QKEYQ'          q.channel of current jobs i/o
        ds.w 0
        dc.w qsend-*
        dc.b 5,'QSEND'          text to pipe/windows queue
        ds.w 0
        dc.w 0
pipn    dc.w 4,'PIPE'           name for test-channel

QKEYQ   bsr.s   krixsv          > return-space, A4:sysv
        move.l  76(a4),a2       maybe the current i/o-queue
        moveq   #-1,d2          flag (to save codespace)
* find the owner channel of a supposed i/o-queue
* with flag D2:=-1: A2:queue  A4:sysvari; else D2=cdt to search for
qky_chn move.l  120(a4),a0      sv.chbas
        move.l  (a0),a5         q.chan 0 always CON
        move.l  4(a5),a5        fetch the handler
        move.w  114(a4),d4      sv.chmax
        move.w  d4,d0           max. existing channels
        lsl.w   #2,d0           adjust to ptrs
        lea     4(a0,d0.w),a0   cdt table ptr top down
qky_lop move.l  -(a0),d0
        bmi.s   qky_nxt         ? channel closed
        move.l  d0,a1
        ror.b   #1,d0
        bmi.s   qky_nxt         ? not an address
        cmp.l   a1,d2           flag -1 or cdt
        beq.s   qky_chx         ? 2nd loop check, cdt found
        cmp.l   a1,a2           lower bound
        bmi.s   qky_nxt         ? queue address below
        move.l  a1,a3           kee cdt
        add.l   (a3),a3         cdt end
        cmp.l   a2,a3           test upper bound
qky_nxt dbpl    d4,qky_lop      ? try next channel
        bmi.s   qky_xit         ? not found
* checking wether PIF took over the i/o-queue
* channel definition table in A1; 1st sorting out safe results
        cmp.l   4(a1),a5        handler
qky_chx beq     qsd_xit         ? is a regular CON/SCR channel
* the only other valid channel is the PIF auxilary window
* we now try to safely sort any other channel out
        moveq   #-2,d4          prepare err.nj
        tst.l   d2              flag
        bgt.s   qky_xit         ? 2nd check channel not found
        moveq   #80,d3
        neg.b   d3              -> byte 176
        cmp.l   (a1),d3         min. cdt-len for CON w. PIF (SCR is 160 bytes)
        bhi.s   qky_xit         ? no PIF-channel found
        add.w   #24,a1
        tst.l   (a1)+           standard sd.x/ymin = PIF sd.x/yhits
        bne.s   qky_xit         ? not the PIF channel if =/= 0
        tst.l   (a1)+           standard sd.x/ysize = PIF sd.x/yhito
        bne.s   qky_xit         ? again
* found PIF present and A1 is not a standard-QL CON-cdt
        move.w  pifdsp(pc),d2   PIF's disp to wdw. parameters
        add.w   d2,a1
        tst.l   (a1)            sd.x/ysize is {0,0} only in PIF aux. wdw.
* (intermediate branch just to save some codespace)
qky_xit bne     qsd_xit         ? still not the aux. channel
* finally we assume(!) the channel being the PIF auxilary window and
* take the last chance fetching that one where the pointer points to
        move.l  144(a5),a2      pt.cchadd pointer's displaced cdt
        tst.b   17(a2)          sd.prwin
        bmi.s   qky_prm         ? primary window
        move.l  -8(a2),a2       true cdt
        add.w   d2,a2           displaced
* we have to search again the table of cdt's to find the channel number
qky_prm sub.w   d2,a2           back to true cdt (and not a queue addr)
        move.l  a2,d2           pass to flag register
        bra.s   qky_chn         - loop again, to find the cdt

* assure parameter return space for one integer
*EXIT:  d1=d5:callers job-id  d2:version string  A0=A4:sysvari
krixsv  moveq   #2,d1           space for integer result
        move.w  $11a,a1         bv.chrix
        jsr     (a1)            > prepare
        moveq   #0,d0           mt_inf
        trap    #1
        move.l  a0,a4           A4: sysvari
        move.l  d1,d5           D5: calling job id
        rts

* send text to a CON-window
QSEND   bsr.s   krixsv          > space, D5:=callers job-id A4:sysvari
        bsr     fchan           > fetch channel no.
        bne.s   qsd_xit         ? none
        add.b   d2,d2           mask off the '#'-flag
        lsr.b   #5,d2           sepa to lo.bits
        subq.b  #4,d2           "!"
        seq     d7              -1
        subq.b  #1,d2           "TO"
        seq     d2
        add.b   d2,d2           -2
        or.b    d2,d7           sepa flag -1:"!" -2:"TO" 0:any other
        bsr.s   quetest         > check wether i/o-queue exists
        bne.s   qsd_xit         ? no, else A2:queue address A1:cdt
        move.l  8(a1),d6        save channel owner
        bsr     fetchst         > fetch text to send
        bne.s   qsd_rrt         ? ret parameter error
        moveq   #-12,d0         prepare err.bn
        move.w  0(a1,a6.l),d4   length
        bmi.s   qsd_srt         ? 32K+, ret err.bn
* fetch and store text from parameterspace to receiver
        addq.b  #1,d7           -> -1:pick 0:wait 1:nowait
        bpl.s   qsd_npk         ? don't pick
        move.l  d6,d1           destination job
        move.l  a0,d5           save ri.len
* pick-channel must be of a non-busy accessible window (the manual is wrong!)
        move.l  48(a6),a0       we try own channel #0, which even during
        move.l  0(a0,a6.l),a0   line interpretion should always be accessible
        moveq   #2,d2           trying a 'wake'
        moveq   #0,d3
        moveq   #124,d0         iop.pick
        trap    #3
        move.l  d5,a0           ri.len back
qsd_npk move.w  224,a4          vector io.qin
        move.w  d4,d2           copy length for remainder
        bra.s   qsd_nxt         - decrement counter and test for 0
qsd_inq addq.l  #1,a1
qsd_inr move.b  1(a1,a6.l),d1
        jsr     (a4)            > store byte
qsd_nxt dbmi    d2,qsd_inq      ? until queue full, or text exhausted
        beq.s   qsd_srt         ? all sent, ret len+1
        tst.b   d7
        bgt.s   qsd_srt         ? early return
        subq.l  #1,a1           retry same char
        tas     $8f(a6)
        bne.s   qsd_inq         ? no <break>, retry
qsd_srt sub.w   d2,d4           ret string position
qsd_rrt add.l   $5c(a6),a0
        move.l  a0,$58(a6)      clear parameterspace
qsd_xit move.l  $58(a6),a1      bv.rip
        subq.l  #2,a1           adjust
        move.l  a1,$58(a6)      for returnspace
        move.w  d4,0(a1,a6.l)   ernum
        moveq   #3,d4           type integer
* sd.extop entry to testing wether a channel defines a window
wdwxtp  moveq   #0,d0           flag/return ok
        rts

* test wether A2 is cdt of a queue owner (CON, PIPE), return A2:=queue header
*ENTRY d1:channel-id  a2:cdt  a4:sysvari
*EXIT  d1/d2:xx d3.w flag -ve:pipe  d4:ernum  a0:xx  a1:cdt a2:queue a4=a4
quetest move.l  d1,a0           channel
        move.l  a2,a1           save cdt address
        moveq   #9,d0           sd.extop
        lea     wdwxtp(pc),a2   routine
        moveq   #0,d3           not waiting
        trap    #3
        move.l  a1,a2           cdt back
* we cannot rely on any buried windows information except the error code
* and as the windows state is unknown, that is it:
        not.l   d0              window if ok or err.nc
        bgt.s   qts_pip         ? else try PIPE definition
* any other channel definition also may include an sd.extop call
* we still don't know for shure wether D1 is a windows channel id.
        moveq   #112,d0         iop.pinf
        trap    #3
* D3 probably must be set to zero for (old?) Atari-QL e-Software
* this can be patched in at address 4 (16-bit value) of the code-file
        move.l  piflag(pc),d3   maybe PIF cdt displacement
        move.l  a0,d1           channel back
        not.l   d0              if ok or with err.nc
        ble.s   qts_pif         ? PIF present & channel is a window
        moveq   #0,d3           clear both, flag and disp
qts_pif move.l  120(a4),a0      sv.chbas
        move.l  (a0)+,a1        cdt of channel 0
* this might not work with some "lightning" versions:
        move.l  4(a1),a0        handler always of CON/SCR type
        cmp.l   4(a2),a0
        move.l  a2,a1           save cdt
        bne.s   qts_pip         ? not a CON/SCR channel, try PIPE
* for currently not accessible (buried) windows we cannot rely
* on the Pointer-I.F. routines. thus, "hand-made":
        tst.l   d3              hi.w was used as a flag
        beq.s   qts_mor         ? no PIF-disp
* with PIF and displacement known the i/o-queue is easily found:
        add.w   d3,a2           add disp for parameter base address
        add.l   -$18+$84(a0),a2 pt.kqoff bears the disp to i/o-queue
        move.l  a2,d2
        ble.s   qts_fld         ? test failed, no queue
        ror.b   #1,d2
        bmi.s   qts_fld         ? also
        move.l  (a2),d2
        ble.s   qts_fld         ? also
        addq.l  #4,a2           true address
qts_fnd subq.l  #4,a2           adjust to queue header
        moveq   #0,d4           'eq' flag "found"
        rts
* if no disp nor flag was found we need some more testing:
qts_mor move.l  (a2),a0         c.d.t. length
        add.l   a2,a0           end of c.d.t.
        add.w   #108,a2         assume queue end-ptr of post-JM QLs
        cmp.l   (a2),a0
        beq.s   qts_fnd         ? standard CON/SCR channel
        cmp.l   -(a2),a0
        beq.s   qts_fnd         ? JM system
qts_fld moveq   #-7,d4          err.nf, maybe SCR, or non-QDOS
qts_ret rts
qts_pip move.l  d1,d3           try linking to a pipe
        lea     pipn(pc),a0     name
        moveq   #-1,d1
        moveq   #1,d0           io.open
        trap    #2
        move.l  d3,d1           channel
        move.l  d0,d4
        bne.s   qts_ret         ? test failed, not a pipe
        moveq   #2,d0           io.close
        trap    #2
        lea     28(a2),a0
        move.l  (a0),d0         output (receiving) queue
        subq.l  #4,a0           ptr to input queue
        bgt.s   qts_opp         ? enable input
        move.l  (a0)+,d0
        beq.s   qts_fld         ? no queue found
qts_opp tst.l   (a0)
        bgt.s   qts_phd         ? other direction already set
        move.l  d0,(a0)         enable pipe bi-directional
qts_phd moveq   #-1,d3          flag 'pipe'
        move.l  a2,a1           cdt
        move.l  d0,a2           queue header address
        moveq   #0,d4
        rts

* fetch a single parameter from basic
*EXIT   d0=d4:ernum  d1.b:-ve on '#'  d3:0 ok. -1 no parm
*       a0:ri-stack space  a1,a6):ptr to parm
* long integer:
fetchli move.w  $118,a1         ca.gtlin
        move.b  1(a3,a6.l),d1   '#'-flag & separator
        bra.s   ftv_num         -
* string:
fetchst move.w  $116,a1         ca.gtstr
ftv_num moveq   #-15,d0         prepare err.bp
        move.l  $58(a6),a0      bv.rip
        sub.l   $5c(a6),a0      bv.ribas
        moveq   #8,d3
        add.l   a3,d3
        cmp.l   d3,a5
        bcs.s   ftv_nix         ? no parameter
        movem.l d1/d6/a0/a2/a5,-(sp) flag, reg, ri-size, top of parm.table
        move.l  d3,a5           upper bound for this call
        jsr     (a1)            > fetch value
        move.l  a5,a3           discard fetched parameter table
        movem.l (sp)+,d1/d6/a0/a2/a5 d1:flag a0:ri.stack a5:parm.top
ftv_nix move.l  d0,d4           preserve and
        bne.s   ftv_ret         ? ret ernum
        subq.w  #1,d3           num. of parameters > 0 'eq' else ret 'mi'
ftv_ret rts

* connect pipe to pipe or CON-type window
PCONNECT bsr    krixsv          > safe returnspace, A4:=sysvari
        bsr.s   fchan           > sending channel
        bne.s   pco_rrt         ? parameter error
        bsr     quetest         >
        bne.s   pco_rrt         ? no queue found
        moveq   #-12,d4         prepare err.bn
        tst.w   d3
        bpl.s   pco_rrt         ? not a PIPE
        moveq   #28,d6          disp to ch.qout
        add.l   a1,d6           save ptr
        bsr.s   fchan           > receiver
        bne.s   pco_rrt         ? parameter error
        bsr     quetest         >
        bne.s   pco_rrt         ? no receiving queue
        move.l  d6,a1
        move.l  a2,(a1)         connect receiver to sender
pco_rrt bra     qsd_xit         - ret ok or error flag

* fetch a qdos/s-basic channel number, 'eq' if found
fchan   bsr.s   fetchli         > fetch channel no.
        bmi.s   qdc_ret         ? parameter error
        move.l  0(a6,a1.l),d7   channel-id
        bmi.s   qdc_ret         ? no
        add.l   $5c(a6),a0
        move.l  a0,$58(a6)      clear parameter stack
* check wether d7 is a valid channel number or a channel-id
*ENTRY d1:'#'-flag  d7:id/qdos-no.  a4:sysvari
*EXIT  d4/flags:ernum  d1:chan-/job-id  d2.b:hash|sepa|use  d7:xx a2:cdt/jdt
        moveq   #-7,d4  prepare err.nf
        move.b  d1,d2   test hash and save sepa-bits
        bpl.s   qdc_qch ? qdos number
        muls    #40,d7
        bmi.s   qdc_ret ? -ve not valid
        add.l   48(a6),d7 bv.chbas
        cmp.l   52(a6),d7 bv.chtop
        bcc.s   qdc_ret ? doesn't exist
        move.l  0(a6,d7.l),d7
qdc_qch move.l  d7,d1
        bmi.s   qdc_ret ? not valid
        cmp.w   114(a4),d7 sv.chmax
        bhi.s   qdc_ret ? doesn't exist
        moveq   #-6,d4  err.no
        move.l  120(a4),a2 sv.chbas
qid_ntr lsl.w   #2,d1   <entry to checking job-id)
        move.l  0(a2,d1.w),d1
        bmi.s   qdc_ret ? closed
        move.l  d1,a2     definition table
        move.l  16(a2),d1 tag to hi.w
        move.w  d7,d1   insert qdos-no.
        cmp.l   d7,d1   check tag no. inclusive
        beq.s   qdc_chn ? channel/job found
        clr.w   d7      clear channel number
        tst.l   d7      ok if parm did not include the tag-no.
        bne.s   qdc_ret ? id doesn't exist
qdc_chn moveq   #0,d4   'eq' if ok
qdc_ret rts

* check wether d7 is a valid job number or a job-id
*ENTRY d1:'#'-flag  d5:calling job  d7:id/qdos-no.  a4:sysvari
*EXIT  d4:ernum  d1:q.no. d7:job-id  a2:jdt
qsdckid moveq   #-2,d4  prepare err.nj
        cmp.w   98(a4),d7 sv.jbmax
        bhi.s   qdc_ret ? job doesn't exist
        moveq   #0,d1
        move.w  d7,d1
        bpl.s   qdc_tjb ?
        move.l  d5,d1   substitute..
        move.l  d5,d7   ..with callers id
qdc_tjb move.l  104(a4),a2 sv.jbbas
        bra.s   qid_ntr -

* find primary window/keybord owner of a job
KEYQ    bsr     krixsv          > space, D5:job A4:sysv
        bsr     fetchli         > fetch job-number/-id
        bne     qsd_rrt         ? parameter error
        move.l  0(a1,a6.l),d7
        bpl.s   kyq_job         ? +ve job number
        move.l  d5,d7           else current job
kyq_job add.l   $5c(a6),a0
        move.l  a0,$58(a6)      clear parameter-stack
        bsr.s   qsdckid         > check id
        bne     qsd_xit         ? not valid
* now: d1:job  a2:jdt  a4:sysvari
        move.l  120(a4),a0      sv.chbas
        move.l  (a0),a1         cdt of channel 0
        move.l  4(a1),d3        always a window handler
        move.w  114(a4),d4      sv.chmax
        move.w  d4,d0
        lsl.w   #2,d0
        lea     4(a0,d0.w),a0   check top down
kyq_lop move.l  -(a0),d0
        bmi.s   kyq_nxt         ? channel closed
        move.l  d0,a1           cdt
        ror.l   #1,d0
        bmi.s   kyq_nxt         ? not an address
        cmp.l   8(a1),d1
        bne.s   kyq_nxt         ? owned by other job
        cmp.l   4(a1),d3        handler
        bne.s   kyq_nxt         ? not a window
        move.l  a1,a2           keep cdt
        move.w  d4,d2           and channel-no.
kyq_nxt dbra    d4,kyq_lop      ? try next channel
* d2/a2 defines the qdos channel of the jobs 1st window. we now
* check wether the PIF is present, and if so find the primary window
        move.w  d2,d4           prepare return value
        move.l  16(a2),d1       channel tag to hi.w
        move.w  d2,d1           and qdos-no.
        move.l  d1,a0           channel id
        moveq   #112,d0         iop.pinf
        moveq   #0,d3           not waiting
        trap    #3
        not.l   d0
        bgt.s   kyq_ret         ? no PIF, return the channel no.
* the next tests depend on the right value for the parameter-table
* offset. as that is not always safe, we return no error but the channel
* number, instead, just in case it may point to the window searched for:
        add.w   pifdsp(pc),a2   ptr to window parameters (re above)
        tst.b   17(a2)          sd.prwin
        bmi.s   kyq_ret         ? this is a primary window
        move.l  -8(a2),d0       primary qdos-cdt
        move.l  d0,a2
        asr.l   #1,d0
        bls.s   kyq_ret         ? not a valid address
        bmi.s   kyq_ret         ? also
        move.l  16(a2),d4       primary's channel tag
        move.w  a2,d4
        sub.w   122(a4),d4      index
        lsr.w   #2,d4           qdos-channel
kyq_ret bra     qsd_xit         - ready (hopefully)

    end  .hpr'96  eof 
